@@ -26,11 +26,8 @@
 #include <libconfig.h>
 #endif
 #ifdef ENABLE_REGEX
-#ifdef LIBPCRE
-#include <pcreposix.h>
-#else
-#include <regex.h>
-#endif
+#define PCRE2_CODE_UNIT_WIDTH 8
+#include <pcre2.h>
 #endif
 
 #ifdef LIBBSD
@@ -86,14 +83,13 @@ static void printcaps(void) {
 static void printsettings(void)
 {
     char buf[NI_MAXHOST];
-    struct addrinfo *a;
     int i;
     struct sslhcfg_protocols_item *p;
     
     for (i = 0; i < cfg.protocols_len; i++ ) {
         p = &cfg.protocols[i];
         fprintf(stderr,
-                "%s addr: %s. libwrap service: %s log_level: %d family %d %d [%s] [%s]\n",
+                "%s addr: %s. libwrap service: %s log_level: %d family %d %d [%s] [%s] [%s]\n",
                 p->name, 
                 sprintaddr(buf, sizeof(buf), p->saddr), 
                 p->service,
@@ -101,80 +97,22 @@ static void printsettings(void)
                 p->saddr->ai_family,
                 p->saddr->ai_addr->sa_family,
                 p->keepalive ? "keepalive" : "",
-                p->fork ? "fork" : "");
-    }
-    fprintf(stderr, "listening on:\n");
-    for (a = addr_listen; a; a = a->ai_next) {
-        fprintf(stderr, 
-                "\t%s\t[%s]\n", 
-                sprintaddr(buf, sizeof(buf), a), 
-                a->ai_flags & SO_KEEPALIVE ? "keepalive" : "");
+                p->fork ? "fork" : "",
+                p->transparent ? "transparent" : ""
+                );
     }
     fprintf(stderr, "timeout: %d\non-timeout: %s\n", cfg.timeout,
             timeout_protocol()->name);
 }
 
 
-/* To be removed in v1.21 */
-const char* ssl_err_msg = "Usage of 'ssl' setting is deprecated and will be removed in v1.21. Please use 'tls' instead\n";
-void ssl_to_tls(char* setting)
-{
-    if (!strcmp(setting, "ssl")) {
-        strcpy(setting, "tls"); /* legacy configuration */
-        log_message(LOG_INFO, ssl_err_msg);
-    }
-}
-
-
-/* Turn 'ssl' command line option to 'tls'. To be removed in v1.21 */
-void cmd_ssl_to_tls(int argc, char* argv[])
-{
-    int i;
-    for (i = 0; i < argc; i++) {
-        if (!strcmp(argv[i], "--ssl")) {
-            strcpy(argv[i], "--tls");
-            /* foreground option not parsed yet, syslog not open, just print on
-             * stderr and hope for the best */
-            fprintf(stderr, "%s", ssl_err_msg);
-        }
-    }
-}
-
-
-/* Extract configuration on addresses and ports on which to listen.
- * out: newly allocated list of addrinfo to listen to
- */
-#ifdef LIBCONFIG
-static int config_resolve_listen(struct addrinfo **listen)
-{
-    int i, res;
-    for (i = 0; i < cfg.listen_len; i++) {
-        res = resolve_split_name(listen, cfg.listen[i].host, cfg.listen[i].port);
-        if (res) return res;
-
-        /* getaddrinfo returned a list of addresses corresponding to the
-         * specification; move the pointer to the end of that list before
-         * processing the next specification, while setting flags for
-         * start_listen_sockets() through ai_flags (which is not meant for
-         * that, but is only used as hint in getaddrinfo, so it's OK) */
-        for (; *listen; listen = &((*listen)->ai_next)) {
-            if (cfg.listen[i].keepalive)
-                (*listen)->ai_flags = SO_KEEPALIVE;
-        }
-    }
-    return 0;
-}
-#endif
-
-
-
-#ifdef LIBCONFIG
 static void setup_regex_probe(struct sslhcfg_protocols_item *p)
+#ifdef ENABLE_REGEX
 {
-    int num_patterns, i, res;
-    regex_t** pattern_list;
-    size_t errsize;
-    char* err;
+    int num_patterns, i, error;
+    pcre2_code** pattern_list;
+    PCRE2_SIZE error_offset;
+    PCRE2_UCHAR8 err_str[120];
 
     num_patterns = p->regex_patterns_len;
 
@@ -183,19 +121,21 @@ static void setup_regex_probe(struct sslhcfg_protocols_item *p)
     p->data = (void*)pattern_list;
 
     for (i = 0; i < num_patterns; i++) {
-        pattern_list[i] = malloc(sizeof(*(pattern_list[i])));
-        CHECK_ALLOC(pattern_list[i], "malloc");
-        res = regcomp(pattern_list[i], p->regex_patterns[i], REG_EXTENDED);
-        if (res) {
-            err = malloc(errsize = regerror(res, pattern_list[i], NULL, 0));
-            CHECK_ALLOC(err, "malloc");
-            regerror(res, pattern_list[i], err, errsize);
-            fprintf(stderr, "%s:%s\n", p->regex_patterns[i], err);
-            free(err);
+        pattern_list[i] = pcre2_compile((PCRE2_SPTR8)p->regex_patterns[i], 
+                                        PCRE2_ZERO_TERMINATED, 0,
+                                        &error, &error_offset, NULL);
+        if (!pattern_list[i]) {
+            pcre2_get_error_message(error, err_str, sizeof(err_str));
+            fprintf(stderr, "compiling pattern /%s/:%d:%s at offset %ld\n",
+                    p->regex_patterns[i], error, err_str, error_offset);
             exit(1);
         }
     }
 }
+#else
+{
+    return;
+}
 #endif
 
 /* For each protocol in the configuration, resolve address and set up protocol
@@ -208,7 +148,7 @@ static void config_protocols()
         struct sslhcfg_protocols_item* p = &(cfg.protocols[i]);
         if (resolve_split_name(&(p->saddr), p->host, p->port)) {
             fprintf(stderr, "cannot resolve %s:%s\n", p->host, p->port);
-            exit(1);
+            exit(4);
         }
 
         p->probe = get_probe(p->name);
@@ -244,7 +184,7 @@ void config_sanity_check(struct sslhcfg_item* cfg) {
 
 /* If compiling with systemd socket support no need to require listen address */
 #ifndef SYSTEMD
-    if (!addr_listen && !cfg->inetd) {
+    if (!cfg->listen_len && !cfg->inetd) {
         fprintf(stderr, "No listening address specified; use at least one -p option\n");
         exit(1);
     }
@@ -258,7 +198,7 @@ int main(int argc, char *argv[], char* envp[])
    extern char *optarg;
    extern int optind;
    int res, num_addr_listen;
-   int *listen_sockets;
+   struct listen_endpoint *listen_sockets;
 
 #ifdef LIBBSD
    setproctitle_init(argc, argv, envp);
@@ -269,8 +209,6 @@ int main(int argc, char *argv[], char* envp[])
    if (res) exit(6);
    if (cfg.verbose > 3)
        sslhcfg_fprint(stderr, &cfg, 0);
-   res = config_resolve_listen(&addr_listen);
-   if (res) exit(4);
    config_protocols();
    config_sanity_check(&cfg);
 
@@ -284,7 +222,7 @@ int main(int argc, char *argv[], char* envp[])
    if (cfg.verbose)
        printsettings();
 
-   num_addr_listen = start_listen_sockets(&listen_sockets, addr_listen);
+   num_addr_listen = start_listen_sockets(&listen_sockets);
 
 #ifdef SYSTEMD
    if (num_addr_listen < 1) {
